# coding: utf-8

from __future__ import absolute_import

from flask import json
from six import BytesIO

from fabric_cf.orchestrator.swagger_server.models.success import Success  # noqa: E501
from fabric_cf.orchestrator.swagger_server.test import BaseTestCase


class TestResourcesController(BaseTestCase):
    """ResourcesController integration test stubs"""

    def test_portalresources_get(self):
        """Test case for portalresources_get

        Retrieve a listing and description of available resources for portal
        """
        query_string = [('graph_format', 'JSON_NODELINK')]
        response = self.client.open(
            '//portalresources',
            method='GET',
            query_string=query_string)
        self.assert200(response,
                       'Response body is : ' + response.data.decode('utf-8'))

    def test_resources_get(self):
        """Test case for resources_get

        Retrieve a listing and description of available resources
        """
        query_string = [('level', 1)]
        response = self.client.open(
            '//resources',
            method='GET',
            query_string=query_string)
        self.assert200(response,
                       'Response body is : ' + response.data.decode('utf-8'))


if __name__ == '__main__':
    import unittest
    unittest.main()



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import bisect

from fabric_cf.actor.core.apis.abc_reservation_mixin import ABCReservationMixin
from fabric_cf.actor.core.common.constants import Constants
from fabric_cf.actor.core.common.exceptions import FrameworkException
from fabric_cf.actor.core.util.reservation_set import ReservationSet


class ReservationSetWrapper:
    """
    Internal representation for all reservations associated with a cycle.
    """
    def __init__(self, *, reservation_set: ReservationSet, cycle: int):
        """
        Constructor
        @params reservation_set: Reservation set
        @params cycle: cycle
        """
        self.reservation_set = reservation_set
        self.cycle = cycle

    def __lt__(self, other):
        if not isinstance(other, ReservationSetWrapper):
            # don't attempt to compare against unrelated types
            return NotImplemented

        return self.cycle < other.cycle

    def __gt__(self, other):
        if not isinstance(other, ReservationSetWrapper):
            # don't attempt to compare against unrelated types
            return NotImplemented

        return self.cycle > other.cycle

    def __str__(self):
        return "{}, cycle: {}".format(self.reservation_set, self.cycle)


class ReservationList:
    """
    Maintains a list of reservations associated with a cycle number. Allows for
    efficient retrieval and removal. Groups reservations by the cycle with which
    they are associated and maintains the reservation sets using a sorted list
    and a hash table. The hash table is used for quick lookup of a given
    reservation set. The sorted list allows for efficient reclaiming of sets,
    which are no longer needed.
    Cost of operations:
    - Insert: first insert to a cycle O(log(cycles)), subsequent inserts to an
    existing cycle: 0(1)
    - Remove: 0(1)
    - Reclaim: 0(k), where k is the number of cycles to reclaim.
    """

    def __init__(self):
        """
        Constructor
        """
        self.count = 0
        # list of ReservationSetWrapper
        self.rset_wrapper_list = []
        # map of <int(cycle), ReservationSet>
        self.cycle_to_rset = {}
        # map of <reservation_id, int(cycle)>
        self.reservation_id_to_cycle = {}

    def add_reservation(self, *, reservation: ABCReservationMixin, cycle: int):
        """
        Adds a reservation associated with a given cycle.
        @params reservation:  the reservation
        @params cycle: the cycle with which to associate the reservation
        """
        if reservation is None or cycle < 0 or reservation.get_reservation_id() is None:
            raise FrameworkException(Constants.INVALID_ARGUMENT)

        if reservation.get_reservation_id() in self.reservation_id_to_cycle:
            existing_cycle = self.reservation_id_to_cycle[reservation.get_reservation_id()]
            if existing_cycle == cycle:
                return
            else:
                raise RuntimeError(
                    "Reservation: #{} is already in the list at a different cycle. Please remove it first, "
                    "before adding to a different cycle".format(reservation.get_reservation_id()))

        reservation_set = None
        if cycle in self.cycle_to_rset:
            reservation_set = self.cycle_to_rset[cycle]
        else:
            reservation_set = ReservationSet()

        if reservation_set.contains(reservation=reservation) is False:
            reservation_set.add(reservation=reservation)
            self.count += 1

        if cycle not in self.cycle_to_rset:
            self.add_to_list(reservation_set=reservation_set, cycle=cycle)
            self.cycle_to_rset[cycle] = reservation_set

        self.reservation_id_to_cycle[reservation.get_reservation_id()] = cycle

    def add_to_list(self, *, reservation_set: ReservationSet, cycle: int):
        """
        Adds an entry to the sorted list
        @params reservation_set: reservation set
        @params cycle: cycle
        """
        entry = ReservationSetWrapper(reservation_set=reservation_set, cycle=cycle)
        bisect.insort(self.rset_wrapper_list, entry)

    def get_all_reservations(self, *, cycle: int) -> ReservationSet:
        """
        Returns all reservations associated with cycles up to and including the specified cycle
        @params cycle : cycle
        @returns a set of reservations associated with the cycles up to and including specified cycle.
        Note that removing from the set will not affect the ReservationList
        """
        result = ReservationSet()
        for entry in self.rset_wrapper_list:
            if entry.cycle <= cycle:
                for reservation in entry.reservation_set.values():
                    result.add(reservation=reservation)
            else:
                break
        return result

    def get_reservations(self, *, cycle: int) -> ReservationSet:
        """
        Returns all reservations associated with the specified cycle.
        @params cycle : cycle
        @return a set of reservations associated with the specified cycle. Note
        that removing from the set will not affect the ReservationList
        """
        result = ReservationSet()

        if cycle in self.cycle_to_rset:
            result = self.cycle_to_rset[cycle].clone()

        return result

    def remove_reservation(self, *, reservation: ABCReservationMixin):
        """
        Removes a reservation from the list.
        @params reservation: reservation to remove
        """
        rid = reservation.get_reservation_id()
        if rid in self.reservation_id_to_cycle:
            cycle = self.reservation_id_to_cycle[rid]
            self.reservation_id_to_cycle.pop(rid)

            if cycle in self.cycle_to_rset:
                reservation_set = self.cycle_to_rset[cycle]
                reservation_set.remove(reservation=reservation)

                if reservation_set.size() == 0:
                    self.cycle_to_rset.pop(cycle)
                self.count -= 1

    def size(self) -> int:
        """
        Returns the number of reservations in this collection
        @returns number of reservations in the collection
        """
        return self.count

    def tick(self, *, cycle: int):
        """
        Removes reservations associated with cycles less than or equal to the given cycle
        @params cycle: cycle
        """
        while True:
            if len(self.rset_wrapper_list) > 0:
                entry = self.rset_wrapper_list[0]
                if entry.cycle <= cycle:
                    self.rset_wrapper_list.remove(entry)

                    if entry.cycle in self.cycle_to_rset:
                        self.cycle_to_rset.pop(entry.cycle)

                    self.count -= entry.reservation_set.size()

                    for reservation in entry.reservation_set.values():
                        self.reservation_id_to_cycle.pop(reservation.get_reservation_id())
                    entry.reservation_set.clear()
                else:
                    break
            else:
                break

    def print(self):
        """
        Print the list (used for tests)
        :return:
        """
        print("rw_list: {} c_to_r: {} r_to_c: {}".format(len(self.rset_wrapper_list), len(self.cycle_to_rset),
                                                         len(self.reservation_id_to_cycle)))
        for rw in self.rset_wrapper_list:
            print("{}".format(rw))

        for k, v in self.cycle_to_rset.items():
            print("{}, cycle: {}".format(v, k))

        for k, v in self.reservation_id_to_cycle.items():
            print("rid: {}, cycle: {}".format(k, v))



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from http.client import INTERNAL_SERVER_ERROR

from fss_utils.http_errors import cors_response

from fabric_cf.orchestrator.core.exceptions import OrchestratorException
from fabric_cf.orchestrator.core.orchestrator_handler import OrchestratorHandler
from fabric_cf.orchestrator.swagger_server.models.success import Success  # noqa: E501
from fabric_cf.orchestrator.swagger_server import received_counter, success_counter, failure_counter
from fabric_cf.orchestrator.swagger_server.response.constants import GET_METHOD, RESOURCES_PATH, PORTAL_RESOURCES_PATH
from fabric_cf.orchestrator.swagger_server.response.utils import get_token


def portalresources_get(graph_format):  # noqa: E501
    """Retrieve a listing and description of available resources for portal

    Retrieve a listing and description of available resources for portal # noqa: E501

    :param graph_format: Graph format
    :type graph_format: str

    :rtype: Success
    """
    handler = OrchestratorHandler()
    logger = handler.get_logger()
    received_counter.labels(GET_METHOD, PORTAL_RESOURCES_PATH).inc()
    try:
        value = handler.portal_list_resources(graph_format_str=graph_format)
        response = Success()
        response.value = value
        success_counter.labels(GET_METHOD, PORTAL_RESOURCES_PATH).inc()
        return response
    except OrchestratorException as e:
        logger.exception(e)
        failure_counter.labels(GET_METHOD, PORTAL_RESOURCES_PATH).inc()
        msg = str(e).replace("\n", "")
        return cors_response(status=e.get_http_error_code(), xerror=str(e), body=msg)
    except Exception as e:
        logger.exception(e)
        failure_counter.labels(GET_METHOD, PORTAL_RESOURCES_PATH).inc()
        msg = str(e).replace("\n", "")
        return cors_response(status=INTERNAL_SERVER_ERROR, xerror=str(e), body=msg)


def resources_get(level: int):  # noqa: E501
    """Retrieve a listing and description of available resources

    :param level: Level of details
    :type level: int
    :param graph_format: Graph format
    :type graph_format: str


    :rtype: Success
    """
    handler = OrchestratorHandler()
    logger = handler.get_logger()
    received_counter.labels(GET_METHOD, RESOURCES_PATH).inc()
    try:
        token = get_token()
        value = handler.list_resources(token=token, level=level)
        response = Success()
        response.value = value
        success_counter.labels(GET_METHOD, RESOURCES_PATH).inc()
        return response
    except OrchestratorException as e:
        logger.exception(e)
        failure_counter.labels(GET_METHOD, RESOURCES_PATH).inc()
        msg = str(e).replace("\n", "")
        return cors_response(status=e.get_http_error_code(), xerror=str(e), body=msg)
    except Exception as e:
        logger.exception(e)
        failure_counter.labels(GET_METHOD, RESOURCES_PATH).inc()
        msg = str(e).replace("\n", "")
        return cors_response(status=INTERNAL_SERVER_ERROR, xerror=str(e), body=msg)




#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from abc import ABC, abstractmethod
from typing import Tuple

import yaml

from fabric_cf.actor.core.plugins.handlers.config_token import ConfigToken


class ConfigurationException(Exception):
    pass


class HandlerBase(ABC):
    @staticmethod
    def load_config(path):
        """
        Read config file
        """
        if path is None:
            raise ConfigurationException("No data source has been specified")
        print("Reading config file: {}".format(path))
        config_dict = None
        with open(path) as f:
            config_dict = yaml.safe_load(f)
        return config_dict

    @abstractmethod
    def create(self, unit: ConfigToken) -> Tuple[dict, ConfigToken]:
        """
        Create
        """

    @abstractmethod
    def modify(self, unit: ConfigToken) -> Tuple[dict, ConfigToken]:
        """
        Create
        """

    @abstractmethod
    def delete(self, unit: ConfigToken) -> Tuple[dict, ConfigToken]:
        """
        Create
        """



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import threading
from typing import List, Tuple, Dict
from http.client import BAD_REQUEST

from fabric_mb.message_bus.messages.lease_reservation_avro import LeaseReservationAvro
from fabric_mb.message_bus.messages.reservation_mng import ReservationMng
from fabric_mb.message_bus.messages.ticket_reservation_avro import TicketReservationAvro
from fabric_mb.message_bus.messages.slice_avro import SliceAvro
from fim.graph.slices.abc_asm import ABCASMPropertyGraph
from fim.slivers.capacities_labels import CapacityHints
from fim.slivers.instance_catalog import InstanceCatalog
from fim.slivers.network_node import NodeSliver
from fim.slivers.network_service import NetworkServiceSliver
from fim.user import ServiceType, ComponentType

from fabric_cf.actor.core.common.constants import Constants
from fabric_cf.actor.core.kernel.reservation_states import ReservationStates
from fabric_cf.actor.fim.fim_helper import FimHelper
from fabric_cf.orchestrator.core.exceptions import OrchestratorException
from fabric_cf.orchestrator.core.reservation_converter import ReservationConverter
from fabric_cf.actor.core.apis.abc_mgmt_controller_mixin import ABCMgmtControllerMixin
from fabric_cf.actor.core.util.id import ID


class OrchestratorSliceWrapper:
    """
    Orchestrator Wrapper Around Slice to hold the computed reservations for processing by Slice Deferred Thread
    """
    def __init__(self, *, controller: ABCMgmtControllerMixin, broker: ID, slice_obj: SliceAvro, logger):
        self.controller = controller
        self.broker = broker
        self.slice_obj = slice_obj
        self.logger = logger
        self.reservation_converter = ReservationConverter(controller=controller, broker=broker)
        self.rid_to_res = {}
        self.computed_reservations = []
        self.demanded = []
        self.thread_lock = threading.Lock()
        self.ignorable_ns = [ServiceType.P4, ServiceType.OVS, ServiceType.MPLS]
        self.supported_ns = [ServiceType.L2STS, ServiceType.L2Bridge, ServiceType.L2PTP]

    def lock(self):
        """
        Lock slice
        :return:
        """
        self.thread_lock.acquire()

    def unlock(self):
        """
        Unlock slice
        :return:
        """
        if self.thread_lock.locked():
            self.thread_lock.release()

    def demanded_reservations(self) -> List[str]:
        """
        Return list of reservations demanded from the Actor/Policy
        @return list of demanded reservations
        """
        return self.demanded

    def mark_demanded(self, *, rid: str):
        """
        Mark a reservation demanded
        @param rid: Reservation id
        """
        if rid not in self.demanded:
            self.demanded.append(rid)

    def get_computed_reservations(self) -> List[TicketReservationAvro]:
        """
        Get computed reservations
        :return: computed reservations
        """
        return self.computed_reservations

    def check_predecessors_ticketed_by_rid(self, *, rid: str) -> bool:
        """
        Check if all the parent reservations have been ticketed
        @param rid: Reservation Id
        @return True if parent reservations are ticketed; False otherwise
        """
        res = self.rid_to_res.get(rid, None)

        if res is not None:
            return self.check_predecessors_ticketed(reservation=res)

        return True

    def check_predecessors_ticketed(self, *, reservation: TicketReservationAvro) -> bool:
        """
        Check if all the parent reservations have been ticketed
        @param reservation: Reservation
        @return True if parent reservations are ticketed; False otherwise
        """
        ret_val = True
        if isinstance(reservation, LeaseReservationAvro) and reservation.get_redeem_predecessors() is not None and \
                len(reservation.get_redeem_predecessors()) > 0:

            state_bin = [ReservationStates.Ticketed, ReservationStates.Failed, ReservationStates.Closed,
                         ReservationStates.CloseWait]

            rids = []
            for r in reservation.get_redeem_predecessors():
                rids.append(r.get_reservation_id())

            reservations = self.controller.get_reservations(rid_list=rids)
            rid_res_map = {}

            for r in reservations:
                rid_res_map[r.get_reservation_id()] = r
                state = ReservationStates(r.get_state())
                self.logger.trace(f"Parent Res# {r.get_reservation_id()} is in state# {state}")

                if state not in state_bin:
                    ret_val = False
                    break

            # Parent reservations have been Ticketed; Update BQM Node and Component Id in Node Map
            # Used by Broker to set vlan - source: (c)
            # local_name source: (a)
            # NSO device name source: (a) - need to find the owner switch of the network service in CBM
            # and take its .name or labels.local_name
            if ret_val:
                sliver = reservation.get_sliver()
                for ifs in sliver.interface_info.interfaces.values():
                    component_name, rid = ifs.get_node_map()
                    parent_res = rid_res_map.get(rid)
                    if parent_res is not None and \
                            ReservationStates(parent_res.get_state()) == ReservationStates.Ticketed:
                        node_sliver = parent_res.get_sliver()
                        component = node_sliver.attached_components_info.get_device(name=component_name)
                        graph_id, bqm_component_id = component.get_node_map()
                        graph_id, node_id = node_sliver.get_node_map()
                        ifs.set_node_map(node_map=(node_id, bqm_component_id))

                        # For shared NICs grab the MAC & VLAN from corresponding Interface Sliver
                        # maintained in the Parent Reservation Sliver
                        if component.get_type() == ComponentType.SharedNIC:
                            parent_res_ifs_sliver = FimHelper.get_site_interface_sliver(component=component,
                                                                                        local_name=ifs.get_labels().local_name)
                            parent_labs = parent_res_ifs_sliver.get_label_allocations()

                            ifs.labels.set_fields(mac=parent_labs.mac, vlan=parent_labs.vlan)

                reservation.set_sliver(sliver=sliver)

        self.logger.trace(f"Res# {reservation.get_reservation_id()} {ret_val}")
        return ret_val

    def get_all_reservations(self) -> List[ReservationMng]:
        """
        Get All reservations
        :return: all reservations
        """
        if self.controller is None:
            return None
        return self.controller.get_reservations(slice_id=ID(uid=self.slice_obj.get_slice_id()))

    def get_slice_name(self) -> str:
        """
        Get Slice name
        :return: slice name
        """
        return self.slice_obj.get_slice_name()

    def get_slice_id(self) -> ID:
        """
        Get Slice Id
        :return: slice id
        """
        return ID(uid=self.slice_obj.get_slice_id())

    def create(self, *, slice_graph: ABCASMPropertyGraph) -> List[TicketReservationAvro]:
        """
        Create a slice
        :param slice_graph: Slice Graph
        :return: List of computed reservations
        """
        try:
            # Build Network Node reservations
            network_node_reservations, node_res_mapping = self.__build_network_node_reservations(slice_graph=slice_graph)

            # Build Network Service reservations
            network_service_reservations = self.__build_network_service_reservations(slice_graph=slice_graph,
                                                                                     node_res_mapping=node_res_mapping)

            # Add Network Node reservations
            for r in network_node_reservations:
                self.controller.add_reservation(reservation=r)

            # Add Network Node reservations
            for r in network_service_reservations:
                self.controller.add_reservation(reservation=r)

            # Add to computed reservations
            for x in network_node_reservations:
                self.computed_reservations.append(x)
                self.rid_to_res[x.get_reservation_id()] = x

            for x in network_service_reservations:
                self.computed_reservations.append(x)
                self.rid_to_res[x.get_reservation_id()] = x

            return self.computed_reservations
        except OrchestratorException as e:
            self.logger.error("Exception occurred while generating reservations for slivers: {}".format(e))
            raise e
        except Exception as e:
            self.logger.error("Exception occurred while generating reservations for slivers: {}".format(e))
            raise OrchestratorException(message=f"Failure to build Slivers: {e}")

    @staticmethod
    def __validate_node_sliver(sliver: NodeSliver):
        """
        Validate Network Node Sliver
        @param sliver Node Sliver
        @raises exception for invalid slivers
        """
        if sliver.get_capacities() is None and sliver.get_capacity_hints() is None:
            raise OrchestratorException(message="Either Capacity or Capacity Hints must be specified!",
                                        http_error_code=BAD_REQUEST)

    @staticmethod
    def __validate_network_service_sliver(sliver: NetworkServiceSliver):
        """
        Validate Network Node Sliver
        @param sliver Node Sliver
        @raises exception for invalid slivers
        """
        for ifs in sliver.interface_info.interfaces.values():
            if ifs.labels is not None:
                vlan = int(ifs.get_labels().vlan)
                if vlan <= Constants.VLAN_START or vlan >= Constants.VLAN_END:
                    raise OrchestratorException(message=f"Allowed range for VLAN ({Constants.VLAN_START}-{Constants.VLAN_END})",
                                                http_error_code=BAD_REQUEST)

    def __build_network_service_reservations(self, slice_graph: ABCASMPropertyGraph,
                                             node_res_mapping: Dict[str, str]) -> List[TicketReservationAvro]:
        """
        Build Network Service Reservations
        @param slice_graph Slice graph
        @param node_res_mapping Mapping of Network Node sliver Id to reservation Id;
                                Needed to add dependency information in network service slivers
        @return list of network service reservations
        """
        reservations = []
        for ns_id in slice_graph.get_all_network_service_nodes():

            # Build Network Service Sliver
            sliver = slice_graph.build_deep_ns_sliver(node_id=ns_id)
            sliver_type = sliver.get_type()

            # Ignore Sliver types P4,OVS and MPLS
            if sliver_type in self.ignorable_ns:
                continue

            # Process only the currently supported Network Sliver types L2STS, L2PTP and L2Bridge
            elif sliver_type in self.supported_ns:

                self.logger.trace(f"Network Service Sliver: {sliver}")

                self.__validate_network_service_sliver(sliver=sliver)

                # Processing Interface Slivers
                if sliver.interface_info is not None:
                    predecessor_reservations = []
                    for ifs in sliver.interface_info.interfaces.values():
                        # Get Mapping information for Interface Sliver from ASM
                        # i.e. [Peer IFS, Peer NS Id, Component Name, Node Id]
                        ifs_mapping = FimHelper.get_interface_sliver_mapping(ifs_node_id=ifs.node_id,
                                                                             slice_graph=slice_graph)

                        if ifs_mapping is None:
                            raise OrchestratorException(message=f"Peer connection point not found for ifs# {ifs}",
                                                        http_error_code=BAD_REQUEST)

                        # capacities (bw in Gbps, burst size is in Mbits) source: (b)
                        # Set Capacities
                        ifs.set_capacities(cap=ifs_mapping.get_peer_ifs().get_capacities())

                        # Set Labels
                        ifs.set_labels(lab=ifs_mapping.get_peer_ifs().get_labels())

                        # Save the parent component name and the parent reservation id in the Node Map
                        parent_res_id = node_res_mapping.get(ifs_mapping.get_node_id(), None)

                        node_map = tuple([ifs_mapping.get_component_name(), parent_res_id])
                        ifs.set_node_map(node_map=node_map)

                        self.logger.trace(f"Interface Sliver: {ifs}")

                        if parent_res_id is not None and parent_res_id not in predecessor_reservations:
                            predecessor_reservations.append(parent_res_id)

                    # Generate reservation for the sliver
                    reservation = self.reservation_converter.generate_reservation(sliver=sliver,
                                                                                  slice_id=self.slice_obj.get_slice_id(),
                                                                                  end_time=self.slice_obj.get_lease_end(),
                                                                                  pred_list=predecessor_reservations)
                    reservations.append(reservation)
            else:

                raise OrchestratorException(message="Not implemented",
                                            http_error_code=BAD_REQUEST)
        return reservations

    def __build_network_node_reservations(self, slice_graph: ABCASMPropertyGraph) \
            -> Tuple[List[TicketReservationAvro], Dict[str, str]]:
        reservations = []
        sliver_to_res_mapping = {}
        for nn_id in slice_graph.get_all_network_nodes():

            # Build Network Node Sliver
            sliver = slice_graph.build_deep_node_sliver(node_id=nn_id)

            # Validate Node Sliver
            self.__validate_node_sliver(sliver=sliver)

            # Compute Requested Capacities from Capacity Hints
            requested_capacities = sliver.get_capacities()
            requested_capacity_hints = sliver.get_capacity_hints()
            catalog = InstanceCatalog()
            if requested_capacities is None and requested_capacity_hints is not None:
                requested_capacities = catalog.get_instance_capacities(
                    instance_type=requested_capacity_hints.instance_type)
                sliver.set_capacities(cap=requested_capacities)

            # Compute Capacity Hints from Requested Capacities
            if requested_capacity_hints is None and requested_capacities is not None:
                instance_type = catalog.map_capacities_to_instance(cap=requested_capacities)
                requested_capacity_hints = CapacityHints().set_fields(instance_type=instance_type)
                sliver.set_capacity_hints(caphint=requested_capacity_hints)

            # Generate reservation for the sliver
            reservation = self.reservation_converter.generate_reservation(sliver=sliver,
                                                                          slice_id=self.slice_obj.get_slice_id(),
                                                                          end_time=self.slice_obj.get_lease_end())
            reservations.append(reservation)

            self.logger.trace(f"Mapped sliver: {sliver.node_id} to res: {reservation.get_reservation_id()}")

            # Maintain Sliver Id to Reservation Mapping
            sliver_to_res_mapping[sliver.node_id] = reservation.get_reservation_id()
        return reservations, sliver_to_res_mapping



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

from abc import abstractmethod
from typing import TYPE_CHECKING

from fabric_cf.actor.core.apis.abc_policy import ABCPolicy


if TYPE_CHECKING:
    from fabric_cf.actor.core.apis.abc_client_reservation import ABCClientReservation
    from fabric_cf.actor.core.util.bids import Bids
    from fabric_cf.actor.core.kernel.resource_set import ResourceSet
    from fabric_cf.actor.core.time.term import Term
    from fabric_cf.actor.core.apis.abc_delegation import ABCDelegation


class ABCClientPolicy(ABCPolicy):
    """
    IClientPolicy defines the policy interface for an actor acting
    as a client of another actor (broker or a orchestrator).
    """
    @abstractmethod
    def demand(self, *, reservation: ABCClientReservation):
        """
        Injects a new resource demand into the demand stream. The reservation
        must be pre-initialized with resource set, term, properties, etc. The
        policy should use this request as an indication that new resources are
        required. The exact mapping of the request to actual requests for
        resources is policy-specific. For example, the policy may choose to
        combine several reservation requests into one, split a reservation
        request onto multiple brokers, etc.

        @params reservation : reservation representing resource demand
        """

    @abstractmethod
    def formulate_bids(self, *, cycle: int) -> Bids:
        """
        Formulates bids to the upstream broker(s). The method should determine
        whether to issue bids in the current cycle. This method should consider
        the current demand, call broker(s) to obtain necessary information and
        decide how to distribute its resource demand. When deciding how to bid,
        also consider any expiring reservation and decide whether to renew and
        adjust their units. The code should only formulate the bids: the actor
        will then issue them.

        Here are some guidelines for implementing this method:
        - Determine the final demand for each resource type.
        - Obtain policy-specific information from upstream brokers.
        - Determine how to split the demand across the brokers and the
        currently renewing reservation.
        - Select candidates to request and renew, and prime them with suggested
        terms, unit counts, and brokers (for new reservations), and specify
        whether the reservation is renewable or not. Set properties as needed for
        e.g., economic bidding.
        - Return a ReservationSet of new reservations to be requested, and a
        ReservationSet of reservations to extend. The returned sets may be empty,
        but not null.

        @params cycle: The current time

        @return Two collections:
                - ticketing - set of new reservations
                - extending - set of reservations to be extended. Can be None if no action should be taken

        @raises Exception in case of error
        """

    @abstractmethod
    def ticket_satisfies(self, *, requested_resources: ResourceSet, actual_resources: ResourceSet,
                         requested_term: Term, actual_term: Term):
        """
        Checks if the resources and term received in a ticket are in compliance
        with what was initially requested. The policy can prevent the application
        of the incoming update if it disagrees with it.

        @params requested_resources: resources requested from broker
        @params actual_resources: resources received from broker
        @params requested_term: term requested from broker
        @params actual_term: term received from broker
        @raises Exception in case of error
        """

    @abstractmethod
    def update_ticket_complete(self, *, reservation: ABCClientReservation):
        """
        Notifies the policy that a ticket update operation has completed. The
        policy may use this upcall to update its internal state.

        @params reservation: reservation for which an update ticket operation has completed
        @raises Exception in case of error
        """

    @abstractmethod
    def update_delegation_complete(self, *, delegation: ABCDelegation):
        """
        Notifies the policy that a delegation update operation has completed. The
        policy may use this upcall to update its internal state.

        @params delegation: delegation for which an update delegation operation has completed
        @raises Exception in case of error
        """



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from fabric_mb.message_bus.messages.result_avro import ResultAvro


class KafkaService:
    def __init__(self):
        from fabric_cf.actor.core.container.globals import GlobalsSingleton
        self.producer = GlobalsSingleton.get().get_kafka_producer()
        self.logger = None

    def set_logger(self, *, logger):
        self.logger = logger
        if self.producer is not None:
            self.producer.set_logger(logger=logger)

    def update_status(self, *, incoming: ResultAvro, outgoing: ResultAvro):
        outgoing.set_code(incoming.get_code())
        outgoing.set_details(incoming.get_details())
        outgoing.set_message(incoming.get_message())
        return outgoing

    def get_first(self, *, result_list: list):
        if result_list is not None and len(result_list) > 0:
            return result_list.__iter__().__next__()
        return None



import connexion
import six

from fabric_cf.orchestrator.swagger_server.models.success import Success  # noqa: E501
from fabric_cf.orchestrator.swagger_server import util
from fabric_cf.orchestrator.swagger_server.response import slices_controller as rc


def slices_create_post(body, slice_name, ssh_key, lease_end_time=None):  # noqa: E501
    """Create slice

    Request to create slice as described in the request. Request would be a graph ML describing the requested resources. Resources may be requested to be created now or in future. On success, one or more slivers are allocated, containing resources satisfying the request, and assigned to the given slice. This API returns list and description of the resources reserved for the slice in the form of Graph ML. Orchestrator would also trigger provisioning of these resources asynchronously on the appropriate sites either now or in the future as requested. Experimenter can invoke get slice API to get the latest state of the requested resources.   # noqa: E501

    :param body:
    :type body: dict | bytes
    :param slice_name: Slice Name
    :type slice_name: str
    :param ssh_key: User SSH Key
    :type ssh_key: str
    :param lease_end_time: Lease End Time for the Slice
    :type lease_end_time: str

    :rtype: Success
    """
    return rc.slices_create_post(body, slice_name, ssh_key, lease_end_time)


def slices_delete_slice_iddelete(slice_id):  # noqa: E501
    """Delete slice.

    Request to delete slice. On success, resources associated with slice or sliver are stopped if necessary, de-provisioned and un-allocated at the respective sites.  # noqa: E501

    :param slice_id: Slice identifier as UUID
    :type slice_id: str

    :rtype: Success
    """
    return rc.slices_delete_slice_iddelete(slice_id)


def slices_get(state):  # noqa: E501
    """Retrieve a listing of user slices

    Retrieve a listing of user slices # noqa: E501

    :param state: Slice state
    :type state: str

    :rtype: Success
    """
    return rc.slices_get(state)


def slices_modify_slice_idput(body, slice_id):  # noqa: E501
    """Modify slice

    Request to modify slice as described in the request. Request would be a Graph ML describing the requested resources for slice or a dictionary for sliver. On success, for one or more slivers are modified. This API returns list and description of the resources reserved for the slice in the form of Graph ML. Orchestrator would also trigger provisioning of the new resources on the appropriate sites either now or in the future based as requested. Modify operations may include add/delete/modify a container/VM/Baremetal server/network or other resources to the slice.  # noqa: E501

    :param body:
    :type body: dict | bytes
    :param slice_id: Slice identifier as UUID
    :type slice_id: str

    :rtype: Success
    """
    return rc.slices_modify_slice_idput(body, slice_id)


def slices_redeem_slice_idpost(slice_id):  # noqa: E501
    """Redeem resources reserved via Create API

    Request that the reserved resources be made provisioned, instantiating or otherwise realizing the resources, such that they have a valid operational status and may possibly be made ready for experimenter use. This operation is synchronous, but may start a longer process, such as creating and imaging a virtual machine.  # noqa: E501

    :param slice_id: Slice identifier as UUID
    :type slice_id: str

    :rtype: Success
    """
    return rc.slices_redeem_slice_idpost(slice_id)


def slices_renew_slice_idpost(slice_id, new_lease_end_time):  # noqa: E501
    """Renew slice

    Request to extend slice be renewed with their expiration extended. If possible, the orchestrator should extend the slivers to the requested expiration time, or to a sooner time if policy limits apply.  # noqa: E501

    :param slice_id: Slice identifier as UUID
    :type slice_id: str
    :param new_lease_end_time: New Lease End Time for the Slice
    :type new_lease_end_time: str

    :rtype: Success
    """
    return rc.slices_renew_slice_idpost(slice_id, new_lease_end_time)


def slices_slice_idget(slice_id):  # noqa: E501
    """slice properties

    Retrieve Slice properties # noqa: E501

    :param slice_id: Slice identifier as UUID
    :type slice_id: str

    :rtype: Success
    """
    return rc.slices_slice_idget(slice_id)


def slices_status_slice_idget(slice_id):  # noqa: E501
    """slice status

    Retrieve the status of a slice. Status would include dynamic reservation or instantiation information. This API is used to provide updates on the state of the resources after the completion of create, which began to asynchronously provision the resources. The response would contain relatively dynamic data, not descriptive data as returned in the Graph ML.  # noqa: E501

    :param slice_id: Slice identifier as UUID
    :type slice_id: str

    :rtype: Success
    """
    return rc.slices_status_slice_idget(slice_id)


#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations
from typing import TYPE_CHECKING

from fabric_cf.actor.core.apis.abc_delegation import ABCDelegation
from fabric_cf.actor.core.apis.abc_reservation_mixin import ABCReservationMixin
from fabric_cf.actor.core.apis.abc_slice import ABCSlice
from fabric_cf.actor.core.core.inventory_slice_manager import InventorySliceManager
from fabric_cf.actor.core.plugins.substrate.substrate_mixin import SubstrateMixin

if TYPE_CHECKING:
    from fabric_cf.actor.core.core.actor import ActorMixin
    from fabric_cf.actor.core.plugins.handlers.handler_processor import HandlerProcessor
    from fabric_cf.actor.core.apis.abc_substrate_database import ABCSubstrateDatabase


class AuthoritySubstrate(SubstrateMixin):
    def __init__(self, *, actor: ActorMixin, db: ABCSubstrateDatabase, handler_processor: HandlerProcessor):
        super().__init__(actor=actor, db=db, handler_processor=handler_processor)
        self.inventory_slice_manager = None
        self.initialized = False

    def __getstate__(self):
        state = self.__dict__.copy()
        del state['logger']
        del state['actor']
        del state['initialized']
        del state['inventory_slice_manager']

        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.logger = None
        self.resource_delegation_factory = None
        self.actor = None

        self.initialized = False
        self.inventory_slice_manager = None

    def initialize(self):
        if not self.initialized:
            super().initialize()
            self.inventory_slice_manager = InventorySliceManager(db=self.get_database(), identity=self.actor,
                                                                 logger=self.get_logger())
            self.initialized = True

    def get_inventory_slice_manager(self) -> InventorySliceManager:
        return self.inventory_slice_manager

    def revisit(self, *, slice_obj: ABCSlice = None, reservation: ABCReservationMixin = None, delegation: ABCDelegation = None):
        if slice_obj is not None and slice_obj.is_inventory():
            self.logger.debug("Recovering inventory slice")
            self.recover_inventory_slice(slice_obj=slice_obj)

    def recover_inventory_slice(self, *, slice_obj: ABCSlice):
        if slice_obj.get_graph_id() is not None:
            self.actor.load_model(graph_id=slice_obj.get_graph_id())


#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from typing import Tuple, List

from fim.slivers.attached_components import AttachedComponentsInfo, ComponentSliver, ComponentType
from fim.slivers.base_sliver import BaseSliver
from fim.slivers.capacities_labels import Capacities, Labels
from fim.slivers.delegations import Delegations, DelegationFormat
from fim.slivers.instance_catalog import InstanceCatalog
from fim.slivers.interface_info import InterfaceSliver
from fim.slivers.network_node import NodeSliver

from fabric_cf.actor.core.apis.abc_reservation_mixin import ABCReservationMixin
from fabric_cf.actor.core.common.constants import Constants
from fabric_cf.actor.core.common.exceptions import BrokerException, ExceptionErrorCode
from fabric_cf.actor.core.policy.inventory_for_type import InventoryForType
from fabric_cf.actor.core.util.id import ID


class NetworkNodeInventory(InventoryForType):
    def __init__(self):
        self.logger = None

    def set_logger(self, *, logger):
        self.logger = logger

    def __getstate__(self):
        state = self.__dict__.copy()
        del state['logger']
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.logger = None

    def __check_capacities(self, *, rid: ID, requested_capacities: Capacities, delegated_capacities: Delegations,
                           existing_reservations: List[ABCReservationMixin]) -> str or None:
        """
        Check if the requested capacities can be satisfied with the available capacities
        :param rid: reservation id of the reservation being served
        :param requested_capacities: Requested Capacities
        :param delegated_capacities: Delegated Capacities
        :param existing_reservations: Existing Reservations served by the same BQM node
        :return: Delegation Id of the delegation which satisfies the request
        :raises: BrokerException in case the request cannot be satisfied
        """
        self.logger.debug(f"requested_capacities: {requested_capacities} for reservation# {rid}")

        delegation_id, deleg = delegated_capacities.get_sole_delegation()
        self.logger.debug(f"Available_capacity_delegations: {deleg} {type(deleg)} format {deleg.get_format()}")
        # ignore pool definitions and references for now
        if deleg.get_format() != DelegationFormat.SinglePool:
            return None
        # get the Capacities object
        delegated_capacity = deleg.get_details()
        # Remove allocated capacities to the reservations
        if existing_reservations is not None:
            for reservation in existing_reservations:
                if rid == reservation.get_reservation_id():
                    continue
                # For Active or Ticketed or Ticketing reservations; reduce the counts from available
                resource_sliver = None
                if reservation.is_ticketing() and reservation.get_approved_resources() is not None:
                    resource_sliver = reservation.get_approved_resources().get_sliver()

                if (reservation.is_active() or reservation.is_ticketed()) and \
                        reservation.get_resources() is not None:
                    resource_sliver = reservation.get_resources().get_sliver()

                if resource_sliver is not None and isinstance(resource_sliver, NodeSliver):
                    self.logger.debug(
                        f"Excluding already assigned resources {resource_sliver.get_capacity_allocations()} to "
                        f"reservation# {reservation.get_reservation_id()}")
                    delegated_capacity = delegated_capacity - resource_sliver.get_capacity_allocations()

        # Compare the requested against available
        delegated_capacity = delegated_capacity - requested_capacities
        negative_fields = delegated_capacity.negative_fields()
        if len(negative_fields) > 0:
            raise BrokerException(error_code=ExceptionErrorCode.INSUFFICIENT_RESOURCES,
                                  msg=f"{negative_fields}")

        return delegation_id

    def __check_shared_nic_labels_and_capacities(self, *, available_component: ComponentSliver,
                                                 requested_component: ComponentSliver) -> ComponentSliver:

        # Check labels
        label_delegations = available_component.get_label_delegations()
        delegation_id, deleg = label_delegations.get_sole_delegation()
        self.logger.debug(f"Available label_delegations: {deleg} {type(deleg)} format {deleg.get_format()}")
        # ignore pool definitions and references for now
        if deleg.get_format() != DelegationFormat.SinglePool:
            return None
        # get the Labels object
        delegated_label = deleg.get_details()

        if delegated_label.bdf is None or len(delegated_label.bdf) < 1:
            message = "No PCI devices available in the delegation"
            self.logger.error(message)
            raise BrokerException(error_code=ExceptionErrorCode.INSUFFICIENT_RESOURCES,
                                  msg=f"{message}")

        labels = Labels()
        labels.set_fields(bdf=delegated_label.bdf[0])
        requested_component.label_allocations = labels

        # Find the VLAN from the BQM Component
        if available_component.network_service_info is None or \
                len(available_component.network_service_info.network_services) != 1:
            message = "Shared NIC Card must have one Network Service"
            self.logger.error(message)
            raise BrokerException(error_code=ExceptionErrorCode.FAILURE,
                                  msg=f"{message}")

        ns_name = next(iter(available_component.network_service_info.network_services))
        ns = available_component.network_service_info.network_services[ns_name]

        if ns.interface_info is None or len(ns.interface_info.interfaces) != 1:
            message = "Shared NIC Card must have one Connection Point"
            self.logger.error(message)
            raise BrokerException(error_code=ExceptionErrorCode.FAILURE,
                                  msg=f"{message}")

        ifs_name = next(iter(ns.interface_info.interfaces))
        ifs = ns.interface_info.interfaces[ifs_name]

        ifs_label_delegations = ifs.get_label_delegations()
        delegation_id, deleg = ifs_label_delegations.get_sole_delegation()
        self.logger.debug(
            f"Available Interface Sliver label_delegations: {deleg} {type(deleg)} format {deleg.get_format()}")
        # ignore pool definitions and references for now
        if deleg.get_format() != DelegationFormat.SinglePool:
            return None

        ifs_delegated_labels = deleg.get_details()
        i = 0
        for pci_id in ifs_delegated_labels.bdf:
            if pci_id == delegated_label.bdf[0]:
                break
            i += 1

        # Updated the Requested component with VLAN, BDF, MAC
        ns_name = next(iter(requested_component.network_service_info.network_services))
        ns = requested_component.network_service_info.network_services[ns_name]
        ifs_name = next(iter(ns.interface_info.interfaces))
        ifs = ns.interface_info.interfaces[ifs_name]

        lab = Labels()
        lab.set_fields(bdf=ifs_delegated_labels.bdf[i], mac=ifs_delegated_labels.mac[i],
                       vlan=ifs_delegated_labels.vlan[i], local_name=ifs_delegated_labels.local_name[i])
        ifs.set_label_allocations(lab=lab)

        self.logger.info(f"Assigned Interface Sliver: {ifs}")
        return requested_component

    def __check_component_labels_and_capacities(self, *, available_component: ComponentSliver, graph_id: str,
                                                requested_component: ComponentSliver) -> ComponentSliver:
        """
        Check if available component capacities, labels to match requested component
        :param available_component: available component
        :param graph_id: BQM graph id
        :param requested_component: requested component
        :return: requested component annotated with properties in case of success, None otherwise
        """
        if requested_component.get_model() is not None and \
                requested_component.get_model() != available_component.get_model():
            return requested_component

        # Checking capacity for component
        capacity_delegations = available_component.get_capacity_delegations()
        delegation_id, deleg = capacity_delegations.get_sole_delegation()
        self.logger.debug(f"Available capacity_delegations: {deleg} {type(deleg)} format {deleg.get_format()}")
        # ignore pool definitions and references for now
        if deleg.get_format() != DelegationFormat.SinglePool:
            return None
        # get the Capacities object
        delegated_capacity = deleg.get_details()
        self.logger.debug(f"available_capacity_delegations : {capacity_delegations} {type(capacity_delegations)} "
                          f"for component {available_component}")

        # Delegated capacity would have been decremented already to exclude allocated shared NICs
        if delegated_capacity.unit < 1:
            message = f"Insufficient Capacities for component: {requested_component}"
            self.logger.error(message)
            raise BrokerException(error_code=ExceptionErrorCode.INSUFFICIENT_RESOURCES,
                                  msg=f"{message}")

        requested_component.capacity_allocations = delegated_capacity

        # Set number of units to 1 explicitly for SharedNIC
        if requested_component.get_type() == ComponentType.SharedNIC:
            requested_component.capacity_allocations.unit = 1

        # Check labels
        # FIXME this isn't quite right - we will need to pick one of available labels, but OK to leave for now
        label_delegations = available_component.get_label_delegations()
        delegation_id, deleg = label_delegations.get_sole_delegation()
        self.logger.debug(f"Available label_delegations: {deleg} {type(deleg)} format {deleg.get_format()}")
        # ignore pool definitions and references for now
        if deleg.get_format() != DelegationFormat.SinglePool:
            return None
        # get the Labels object
        delegated_label = deleg.get_details()

        if requested_component.get_type() == ComponentType.SharedNIC:
            requested_component = self.__check_shared_nic_labels_and_capacities(available_component=available_component,
                                                                                requested_component=requested_component)
        else:
            requested_component.label_allocations = delegated_label

        node_map = tuple([graph_id, available_component.node_id])
        requested_component.set_node_map(node_map=node_map)

        return requested_component

    def __exclude_allocated_pci_device_from_shared_nic(self, shared_nic: ComponentSliver,
                                                       allocated_nic: ComponentSliver) -> ComponentSliver:
        """
        For Shared NIC cards, exclude the already assigned PCI addresses from the available PCI addresses in
        BQM Component Sliver for the NIC Card
        """

        if shared_nic.get_type() != ComponentType.SharedNIC and allocated_nic.get_type() != ComponentType.SharedNIC:
            raise BrokerException(error_code=ExceptionErrorCode.INVALID_ARGUMENT,
                                  msg=f"shared_nic: {shared_nic} allocated_nic: {allocated_nic}")

        # Reduce capacity for component
        capacity_delegations = shared_nic.get_capacity_delegations()
        delegation_id, deleg = capacity_delegations.get_sole_delegation()
        # ignore pool definitions and references for now
        if deleg.get_format() != DelegationFormat.SinglePool:
            return None
        # get the Capacities object
        delegated_capacity = deleg.get_details()

        # Exclude already allocated Shared NIC cards
        delegated_capacity -= allocated_nic.get_capacity_allocations()

        self.logger.debug(f"Allocated NIC: {allocated_nic} labels: {allocated_nic.get_label_allocations()}")

        # Get the Allocated PCI address
        allocated_labels = allocated_nic.get_label_allocations()

        label_delegations = shared_nic.get_label_delegations()
        delegation_id, deleg = label_delegations.get_sole_delegation()

        # ignore pool definitions and references for now
        if deleg.get_format() != DelegationFormat.SinglePool:
            return None
        # get the Labels object
        delegated_label = deleg.get_details()

        # Remove allocated PCI address from delegations
        excluded_labels = []

        if isinstance(allocated_labels.bdf, list):
            excluded_labels = allocated_labels.bdf
        else:
            excluded_labels = [allocated_labels.bdf]

        for e in excluded_labels:
            self.logger.debug(f"Excluding PCI device {e}")
            delegated_label.bdf.remove(e)

        return shared_nic

    def __check_components(self, *, rid: ID, requested_components: AttachedComponentsInfo, graph_id: str,
                           graph_node: NodeSliver, existing_reservations: List[ABCReservationMixin]) -> AttachedComponentsInfo:
        """
        Check if the requested capacities can be satisfied with the available capacities
        :param rid: reservation id of the reservation being served
        :param requested_components: Requested components
        :param graph_id: BQM graph id
        :param graph_node: BQM graph node identified to serve the reservation
        :param existing_reservations: Existing Reservations served by the same BQM node
        :return: Components updated with the corresponding BQM node ids
        :raises: BrokerException in case the request cannot be satisfied
        """

        self.logger.debug(f"requested_components: {requested_components} for reservation# {rid}")
        for name, requested_component in requested_components.devices.items():
            resource_type = requested_component.get_type()
            available_components = graph_node.attached_components_info.get_devices_by_type(resource_type=resource_type)

            self.logger.debug(f"Resource Type: {resource_type} available_components: {available_components}")

            if available_components is None or len(available_components) == 0:
                raise BrokerException(error_code=ExceptionErrorCode.INSUFFICIENT_RESOURCES,
                                      msg=f"Component of type: {resource_type} not available in "
                                          f"graph node: {graph_node.node_id}")

            for reservation in existing_reservations:
                if rid == reservation.get_reservation_id():
                    continue
                # For Active or Ticketed or Ticketing reservations; reduce the counts from available
                allocated_sliver = None
                if reservation.is_ticketing() and reservation.get_approved_resources() is not None:
                    allocated_sliver = reservation.get_approved_resources().get_sliver()

                if (reservation.is_active() or reservation.is_ticketed()) and reservation.get_resources() is not None:
                    allocated_sliver = reservation.get_resources().get_sliver()

                if allocated_sliver is not None and isinstance(allocated_sliver, NodeSliver) and \
                        allocated_sliver.attached_components_info is not None:

                    for allocated in allocated_sliver.attached_components_info.devices.values():
                        if allocated.get_node_map() is not None:
                            self.logger.debug(f"Already allocated components {allocated} of resource_type "
                                              f"{resource_type} to reservation# {reservation.get_reservation_id()}")

                            for av in available_components:
                                node_map = allocated.get_node_map()

                                if node_map is not None and av.node_id == node_map[1]:

                                    if av.get_type() == ComponentType.SharedNIC:
                                        self.logger.debug(f"Excluding component {allocated} assigned to "
                                                          f"res# {reservation.get_reservation_id()}")
                                        # Exclude already allocated PCI address from available PCI devices
                                        av = self.__exclude_allocated_pci_device_from_shared_nic(shared_nic=av,
                                                                                                 allocated_nic=allocated)

                                    else:
                                        self.logger.debug(f"Excluding component {allocated} assigned to "
                                                          f"res# {reservation.get_reservation_id()}")
                                        # Exclude already allocated component from available components
                                        graph_node.attached_components_info.remove_device(av.get_name())
                                    break

            available_components = graph_node.attached_components_info.get_devices_by_type(resource_type=resource_type)
            self.logger.debug(f"available_components after excluding allocated components: {available_components}")

            if available_components is None or len(available_components) == 0:
                raise BrokerException(error_code=ExceptionErrorCode.INSUFFICIENT_RESOURCES,
                                      msg=f"Component of type: {resource_type} not available in "
                                          f"graph node: {graph_node.node_id}")

            for component in available_components:
                # check model matches the requested model
                requested_component = self.__check_component_labels_and_capacities(available_component=component,
                                                                                   graph_id=graph_id,
                                                                                   requested_component=requested_component)
                if requested_component.get_node_map() is not None:
                    self.logger.info(f"Found a matching component with resource model "
                                     f"{requested_component.get_model()}")

                    self.logger.info(f"Assigning {component.node_id} to component# "
                                     f"{requested_component} in reservation# {rid} ")

                    # Remove the component from available components as it is assigned
                    graph_node.attached_components_info.remove_device(component.get_name())
                    break

            if requested_component.get_node_map() is None:
                raise BrokerException(error_code=ExceptionErrorCode.INSUFFICIENT_RESOURCES,
                                      msg=f"Component of type: {resource_type} not available in "
                                          f"graph node: {graph_node.node_id}")

        return requested_components

    def allocate(self, *, rid: ID, requested_sliver: BaseSliver, graph_id: str, graph_node: BaseSliver,
                 existing_reservations: List[ABCReservationMixin]) -> Tuple[str, BaseSliver]:
        """
        Allocate an extending or ticketing reservation
        :param rid: reservation id of the reservation to be allocated
        :param requested_sliver: requested sliver
        :param graph_id: BQM graph id
        :param graph_node: BQM graph node identified to serve the reservation
        :param existing_reservations: Existing Reservations served by the same BQM node
        :return: Tuple of Delegation Id and the Requested Sliver annotated with BQM Node Id and other properties
        :raises: BrokerException in case the request cannot be satisfied
        """
        if graph_node.get_capacity_delegations() is None or rid is None:
            raise BrokerException(error_code=Constants.INVALID_ARGUMENT,
                                  msg=f"capacity_delegations is missing or reservation is None")

        if not isinstance(requested_sliver, NodeSliver):
            raise BrokerException(error_code=Constants.INVALID_ARGUMENT,
                                  msg=f"resource type: {requested_sliver.get_type()}")

        if not isinstance(graph_node, NodeSliver):
            raise BrokerException(error_code=Constants.INVALID_ARGUMENT,
                                  msg=f"resource type: {graph_node.get_type()}")

        # Always use requested capacities to be mapped from flavor i.e. capacity hints
        requested_capacity_hints = requested_sliver.get_capacity_hints()
        catalog = InstanceCatalog()
        requested_capacities = catalog.get_instance_capacities(instance_type=requested_capacity_hints.instance_type)

        # Check if Capacities can be satisfied
        delegation_id = self.__check_capacities(rid=rid,
                                                requested_capacities=requested_capacities,
                                                delegated_capacities=graph_node.get_capacity_delegations(),
                                                existing_reservations=existing_reservations)

        # Check if Components can be allocated
        if requested_sliver.attached_components_info is not None:
            requested_sliver.attached_components_info = self.__check_components(
                rid=rid,
                requested_components=requested_sliver.attached_components_info,
                graph_id=graph_id,
                graph_node=graph_node,
                existing_reservations=existing_reservations)

        requested_sliver.capacity_allocations = requested_capacities
        requested_sliver.label_allocations = Labels().set_fields(instance_parent=graph_node.get_name())

        requested_sliver.set_node_map(node_map=(graph_id, graph_node.node_id))

        self.logger.info(f"Reservation# {rid} is being served by delegation# {delegation_id} "
                         f"node# [{graph_id}/{graph_node.node_id}]")

        return delegation_id, requested_sliver

    def free(self, *, count: int, request: dict = None, resource: dict = None) -> dict:
        return



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import threading
from datetime import datetime

from fabric_cf.actor.core.apis.abc_authority_reservation import ABCAuthorityReservation
from fabric_cf.actor.core.apis.abc_reservation_mixin import ABCReservationMixin
from fabric_cf.actor.core.apis.abc_server_reservation import ABCServerReservation
from fabric_cf.actor.core.time.actor_clock import ActorClock
from fabric_cf.actor.core.time.calendar.base_calendar import BaseCalendar
from fabric_cf.actor.core.util.reservation_holdings import ReservationHoldings
from fabric_cf.actor.core.util.reservation_list import ReservationList
from fabric_cf.actor.core.util.reservation_set import ReservationSet


class AuthorityCalendar(BaseCalendar):
    """
    An AuthorityCalendar is used to organized reservation information for an authority. It extends the functionality of
    BaseCalendar with a number of collections:
    - requests: a collection of client requests organized by the time to be serviced
    - closing: a collection of client reservations organized by closing time
    - outlays: a collection of active reservations (outlays)
    """

    def __init__(self, *, clock: ActorClock):
        """
        Creates a new instance.
        @params clock : clock factory
        """
        # List of incoming requests grouped by start cycle.
        super().__init__(clock=clock)
        self.requests = ReservationList()
        # List of reservations to be closed grouped by cycle.
        self.closing = ReservationList()
        # All currently active reservations.
        self.outlays = ReservationHoldings()
        self.lock = threading.Lock()

    def __getstate__(self):
        state = self.__dict__.copy()
        del state['lock']
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.lock = threading.Lock()

    def remove(self, *, reservation: ABCReservationMixin):
        """
        Removes the specified reservation from the calendar.
        @params reservation : reservation to remove
        """
        if isinstance(reservation, ABCServerReservation):
            self.remove_request(reservation=reservation)
            self.remove_closing(reservation=reservation)

        if isinstance(reservation, ABCAuthorityReservation):
            self.remove_outlay(reservation=reservation)

    def remove_schedule_or_in_progress(self, *, reservation: ABCReservationMixin):
        """
        Removes the specified reservations from all internal calendar data
        structures that represent operations to be scheduled in the future or
        operations that are currently in progress. Does not remove the
        reservation from the outlays list
        @params reservation: reservation to remove
        """
        if isinstance(reservation, ABCServerReservation):
            self.remove_request(reservation=reservation)
            self.remove_closing(reservation=reservation)

    def get_requests(self, *, cycle: int) -> ReservationSet:
        """
        Returns all client requests for the specified cycle.
        @params cycle:  cycle
        @returns set of requests for the cycle
        """
        try:
            self.lock.acquire()
            return self.requests.get_reservations(cycle=cycle)
        finally:
            self.lock.release()

    def add_request(self, *, reservation: ABCReservationMixin, cycle: int):
        """
        Adds a new client request.
        @params reservation: reservation to add
        @params cycle: cycle
        """
        try:
            self.lock.acquire()
            self.requests.add_reservation(reservation=reservation, cycle=cycle)
        finally:
            self.lock.release()

    def remove_request(self, *, reservation: ABCReservationMixin):
        """
        Removes the specified reservation from the request list.
        @params reservation: reservation to remove
        """
        try:
            self.lock.acquire()
            self.requests.remove_reservation(reservation=reservation)
        finally:
            self.lock.release()

    def get_closing(self, *, cycle: int) -> ReservationSet:
        """
        Returns all reservations scheduled for closing at the specified cycle.
        @params cycle: cycle
        @returns set of reservations scheduled for closing at the cycle
        """
        try:
            self.lock.acquire()
            result = self.closing.get_all_reservations(cycle=cycle)
            return result
        finally:
            self.lock.release()

    def add_closing(self, *, reservation: ABCReservationMixin, cycle: int):
        """
        Adds a reservation to the closing list.
        @params reservation: reservation to add
        @params cycle: cycle
        """
        try:
            self.lock.acquire()
            self.closing.add_reservation(reservation=reservation, cycle=cycle)
        finally:
            self.lock.release()

    def remove_closing(self, *, reservation: ABCReservationMixin):
        """
        Removes the specified reservation from the closing list.
        @params reservation: reservation to remove
        """
        try:
            self.lock.acquire()
            self.closing.remove_reservation(reservation=reservation)
        finally:
            self.lock.release()

    def add_outlay(self, *, reservation: ABCReservationMixin, start: datetime, end: datetime):
        """
        Adds an allocated client reservation.
        @params reservation: reservation to add
        @params start: start time
        @params end: end time
        """
        try:
            self.lock.acquire()
            self.outlays.add_reservation(reservation=reservation, start=ActorClock.to_milliseconds(when=start),
                                         end=ActorClock.to_milliseconds(when=end))
        finally:
            self.lock.release()

    def remove_outlay(self, *, reservation: ABCReservationMixin):
        """
        Removes a reservation from the outlays list.
        @params reservation: reservation to remove
        """
        try:
            self.lock.acquire()
            self.outlays.remove_reservation(reservation=reservation)
        finally:
            self.lock.release()

    def get_outlays(self, *, d: datetime = None) -> ReservationSet:
        """
        Returns the active client reservations.
        @returns set of all active client reservations
        """
        try:
            self.lock.acquire()
            if d is None:
                return self.outlays.get_reservations()
            else:
                return self.outlays.get_reservations(time=ActorClock.to_milliseconds(when=d))
        finally:
            self.lock.release()

    def tick(self, *, cycle: int):
        try:
            self.lock.acquire()
            super().tick(cycle=cycle)
            self.requests.tick(cycle=cycle)
            self.closing.tick(cycle=cycle)

            ms = self.clock.cycle_end_in_millis(cycle=cycle)
            self.outlays.tick(time=ms)
        finally:
            self.lock.release()



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

from abc import abstractmethod, ABC
from datetime import datetime
from typing import TYPE_CHECKING, Tuple

if TYPE_CHECKING:
    from fim.graph.abc_property_graph import ABCPropertyGraph

    from fabric_cf.actor.core.kernel.slice import SliceTypes
    from fabric_cf.actor.security.auth_token import AuthToken
    from fabric_cf.actor.core.util.resource_type import ResourceType
    from fabric_cf.actor.core.util.id import ID
    from fabric_cf.actor.core.kernel.slice_state_machine import SliceState, SliceOperation


class ABCSlice(ABC):
    """
    ISlice describes the programming interface to a slice object. Each slice has a name (not necessarily unique)
    and a globally unique identifier. Slices are used to organize groups of reservations. Each reservation belongs
    to exactly one slice. Slices information is passed to upstream actors as part of ticket and lease requests.
    There are several slice types:
        inventory slices: are used to organized reservations that represent an inventory.
                          For example, allocated resources, or resources to be used to satisfy client requests
        client slices: are used on server actors to group reservations representing client requests
                       (allocated/assigned resources).
        broker client slices: are client slices that represent the requests from a broker that acts as a
                              client of the containing actor

    Each slice contains a number of properties lists, which can be used to store properties applicable to all
    reservations associated with the slice. Properties defined in the slice are automatically inherited by reservations.
    Each reservation can also override a property inherited by the slice, but defining it in its appropriate
    properties list.
    """

    @abstractmethod
    def get_config_properties(self) -> dict:
        """
        Returns the slice configuration properties list

        Returns:
            configuration properties list
        """

    @abstractmethod
    def get_description(self) -> str:
        """
        Returns the slice description

        Returns:
            slice description
        """

    @abstractmethod
    def get_name(self) -> str:
        """
        Returns the slice name

        Returns:
            slice name
        """

    @abstractmethod
    def get_owner(self) -> AuthToken:
        """
        Returns the slice owner

        Returns:
            slice owner
        """

    @abstractmethod
    def get_resource_type(self) -> ResourceType:
        """
        Returns the resource type of the slice (if any).

        Returns:
            slice resource type
        """

    @abstractmethod
    def get_slice_id(self) -> ID:
        """
        Returns the slice id.

        Returns:
            slice id
        """

    @abstractmethod
    def is_broker_client(self) -> bool:
        """
        Checks if the slice is a broker client slice (a client slice within an authority that represents a broker).

        Returns:
            true if the slice is a broker client slice
        """

    @abstractmethod
    def is_client(self) -> bool:
        """
        Checks if the slice is a client slice.

        Returns:
            true if the slice is a client slice
        """

    @abstractmethod
    def is_inventory(self) -> bool:
        """
        Checks if the slice is a inventory slice.

        Returns:
            true if the slice is a inventory slice
        """

    @abstractmethod
    def set_broker_client(self):
        """
        Marks the slice as a broker client slice (a client slice within an authority that represents a broker).
        """

    @abstractmethod
    def set_client(self):
        """
        Marks the slice as a client slice.
        """

    @abstractmethod
    def set_inventory(self, *, value: bool):
        """
        Sets the inventory flag.

        Args:
            value: inventory status: true, inventory slice, false, client slice
        """

    @abstractmethod
    def set_description(self, *, description: str):
        """
        Sets the slice description.

        Args:
            description: description
        """

    @abstractmethod
    def set_name(self, *, name: str):
        """
        Sets the slice name.

        Args:
            name: name
        """

    @abstractmethod
    def set_owner(self, *, owner: AuthToken):
        """
        Sets the slice owner.

        Args:
            owner: owner
        """

    @abstractmethod
    def set_resource_type(self, *, resource_type: ResourceType):
        """
        Sets the resource type.

        Args:
            resource_type: resource type
        """

    @abstractmethod
    def clone_request(self):
        """
        Makes a minimal clone of the slice object sufficient for
        cross-actor calls.

        @return a slice object to use when making cross-actor calls.
        """

    @abstractmethod
    def set_config_properties(self, *, value: dict):
        """
        Set handlers properties
        @param value: value
        """

    @abstractmethod
    def get_slice_type(self) -> SliceTypes:
        """
        Return slice type
        """

    @abstractmethod
    def set_graph(self, *, graph: ABCPropertyGraph):
        """
        Sets the resource model graph.

        @param graph graph
        """

    @abstractmethod
    def get_graph(self) -> ABCPropertyGraph:
        """
        Gets the resource model graph.
        """

    @abstractmethod
    def set_graph_id(self, graph_id: str):
        """
        Set graph id
        @param graph_id:  graph_id
        """

    @abstractmethod
    def get_graph_id(self) -> str:
        """
        Returns the graph id

        @return graph id
        """

    @abstractmethod
    def get_state(self) -> SliceState:
        """
        Return Slice State
        @return slice state
        """

    @abstractmethod
    def transition_slice(self, *, operation: SliceOperation) -> Tuple[bool, SliceState]:
        """
        Attempt to transition a slice to a new state
        @param operation slice operation
        @return Slice State
        @throws Exception in case of error
        """

    @abstractmethod
    def is_dirty(self) -> bool:
        """
        Checks if the slice has uncommitted updates.

        Returns:
            true if the slice has an uncommitted updates
        """

    @abstractmethod
    def set_dirty(self):
        """
        Marks the slice as containing uncommitted updates.
        """

    @abstractmethod
    def clear_dirty(self):
        """
        Marks that the slice has no uncommitted updates or state transitions.
        """

    @abstractmethod
    def is_stable_ok(self) -> bool:
        """
        Is Slice in state Stable OK
        @return True if Slice state is Stable OK, false otherwise
        """

    def is_stable_error(self) -> bool:
        """
        Is Slice in state Stable Error
        @return True if Slice state is Stable Error, false otherwise
        """

    def is_stable(self) -> bool:
        """
        Is Slice in state StableOK/StableError
        @return True if Slice state is StableOK/StableError, false otherwise
        """

    def is_dead_or_closing(self) -> bool:
        """
        Is Slice in state Dead/Closing
        @return True if Slice state is Dead/Closing, false otherwise
        """

    def is_dead(self) -> bool:
        """
        Is Slice in state Dead
        @return True if Slice state is Dead, false otherwise
        """

    def get_lease_end(self) -> datetime:
        """
        Return lease end time
        """

    def get_lease_start(self) -> datetime:
        """
        Return lease start time
        """

    def set_lease_end(self, *, lease_end: datetime):
        """
        Return lease end time
        """

    def set_lease_start(self, *, lease_start: datetime):
        """
        Return lease start time
        """



import json
import traceback

from fss_utils.jwt_manager import ValidateCode

from fabric_cf.actor.core.common.constants import Constants


class TokenException(Exception):
    """
    Token exception
    """


class FabricToken:
    """
    Represents the Fabric Token issues by Credential Manager
    """
    def __init__(self, *, token: str, logger):
        if token is None:
            raise TokenException('Token: {} is None'.format(token))

        self.logger = logger
        self.encoded_token = token
        self.decoded_token = None

    def get_encoded_token(self) -> str:
        """
        Get Encoded token string
        @return encoded token
        """
        return self.encoded_token

    def get_decoded_token(self) -> dict:
        """
        Get Decoded token
        @return Decoded token
        """
        if self.decoded_token is None:
            self.validate()
        return self.decoded_token

    def validate(self) -> dict:
        """
        Validate the token
        @raise Exception in case of error
        """
        try:
            # validate the token
            from fabric_cf.actor.core.container.globals import GlobalsSingleton
            jwt_validator = GlobalsSingleton.get().get_jwt_validator()
            verify_exp = GlobalsSingleton.get().get_config().get_oauth_config().get(
                Constants.PROPERTY_CONF_O_AUTH_VERIFY_EXP, True)

            if jwt_validator is not None:
                self.logger.info("Validating CI Logon token")
                code, token_or_exception = jwt_validator.validate_jwt(token=self.encoded_token, verify_exp=verify_exp)
                if code is not ValidateCode.VALID:
                    self.logger.error(f"Unable to validate provided token: {code}/{token_or_exception}")
                    raise TokenException(f"Unable to validate provided token: {code}/{token_or_exception}")
            else:
                raise TokenException("JWT Token validator not initialized, skipping validation")

            self.decoded_token = token_or_exception
            self.logger.debug(json.dumps(self.decoded_token))

            return self.decoded_token
        except Exception as e:
            self.logger.error(traceback.format_exc())
            self.logger.error("Exception occurred while validating the token e: {}".format(e))
            raise e
